"""Serial port transport layer, uses RTS/CTS for flow-control"""
import serial as pyserial
import threading
import string
import binascii
import time
import sys
import select
from baseclass import transports_base

# basically a wrapper for Serial
class transports_rs232(transports_base):
    def __init__(self, port, *args, **kwargs):
        """Initializes a serial transport, requires open serial port and message callback as arguments"""
        super(transports_rs232, self).__init__(*args, **kwargs)
        self.line_terminator = "\r\n"
        self._terminator_slice = -1*len(self.line_terminator)
        # For tracking state changes
        self._previous_states = {
            'getCTS': None,
            'getDSR': None,
            'getRI': None,
            'getCD': None
        }
        self._current_states = {
            'getCTS': None,
            'getDSR': None,
            'getRI': None,
            'getCD': None
        }
        self.print_debug = False
        self.serial_port = port
        self.initialize_serial()

    def initialize_serial(self):
        """Creates a background thread for reading the serial port"""
        self.input_buffer = ""
        self.receiver_thread = threading.Thread(target=self.serial_reader)
        self.receiver_thread.setDaemon(1)
        self.receiver_thread.start()

    def serial_reader(self):
        self.serial_alive = True
        if self.serial_port.rtscts:
            self.serial_port.setRTS(True)
        try:
            while self.serial_alive:
                for method in self._current_states:
                    self._current_states[method] = getattr(self.serial_port, method)()
                    if self._current_states[method] != self._previous_states[method]:
                        print " *** %s changed to %d *** " % (method, self._current_states[method])
                        self._previous_states[method] = self._current_states[method]
                rd, wd, ed  = select.select([ self.serial_port, ], [], [ self.serial_port, ], 5) # Wait up to 5s for new data
                if not self.serial_port.inWaiting():
                    # Don't try to read if there is no data, instead sleep (yield) a bit
                    time.sleep(0)
                    continue
                data = self.serial_port.read(1)
                if len(data) == 0:
                    continue
                if self.print_debug:
                    # hex-encode unprintable characters
                    #if data not in string.letters.join(string.digits).join(string.punctuation).join("\r\n"):
                    #     sys.stdout.write("\\0x".join(binascii.hexlify(data)))
                    # OTOH repr was better afterall
                    if data not in self.line_terminator:
                        sys.stdout.write(repr(data))
                    else:
                        sys.stdout.write(data)
                # Put the data into inpit buffer and check for CRLF
                self.input_buffer += data
                # Trim prefix NULLs and linebreaks
                self.input_buffer = self.input_buffer.lstrip(chr(0x0) + self.line_terminator)
                #print "input_buffer=%s" % repr(self.input_buffer)
                if (    len(self.input_buffer) > 0
                    and self.input_buffer[self._terminator_slice:] == self.line_terminator):
                    # Got a message, parse it (sans the CRLF) and empty the buffer
                    #print "DEBUG: calling self.message_received()"
                    self.message_received(self.input_buffer[:self._terminator_slice])
                    self.input_buffer = ""

#        except (IOError, pyserial.SerialException), e:
# something overwrites the module when running I get <type 'exceptions.AttributeError'>: 'NoneType' object has no attribute 'SerialException' if port fails...
        except (IOError), e:
            print "Got exception %s" % e
            self.serial_alive = False
            # It seems we cannot really call this from here, how to detect the problem in main thread ??
            #self.launcher_instance.unload_device(self.object_name)

    def abort_command(self):
        """Uses the break-command to issue "Device clear", from the SCPI documentation (for HP6632B): The status registers, the error queue, and all configuration states are left unchanged when a device clear message is received. Device clear performs the following actions:
 - The input and output buffers of the dc source are cleared.
 - The dc source is prepared to accept a new command string."""
        self.serial_port.sendBreak()

    def stop_serial(self):
        """Stops the serial port thread and closes the port"""
        self.serial_alive = False
        self.receiver_thread.join()
        self.serial_port.close()

    def quit(self):
        """Shuts down any background threads that might be active"""
        self.stop_serial()

    def incoming_data(self):
        """It seems there is no better way to check for transaction-in-progress than this (I was hoping RI or some other modem signal would be used)"""
        return bool(self.serial_port.inWaiting())

    def send_command(self, command):
        """Adds the line terminator and writes the command out"""
        if self.serial_port.rtscts:
            while not self.serial_port.getCTS():
                # Yield while waiting for CTS
                time.sleep(0)
        send_str = command + self.line_terminator
        self.serial_port.write(send_str)



"""HP/Agilent 3362B specific device implementation and helpers"""

#import os,sys
## Add the parent dir to search paths
#libs_dir = os.path.join(os.path.dirname( os.path.realpath( __file__ ) ),  '..',)
#if os.path.isdir(libs_dir):                                       
#    sys.path.append(libs_dir)

from scpi import scpi_device

class hp6632b(scpi_device):
    """Adds the HP/Agilent 3362B specific SCPI commands as methods"""

    def __init__(self, transport, *args, **kwargs):
        """Initializes a device for the given transport"""
        super(hp6632b, self).__init__(transport, *args, **kwargs)
        self.scpi.ask_default_wait = 0.050 # Average aquisition time is 30ms + 20ms processing time

    def set_low_current_mode(self, state):
        """The low-current mode is enabled by setting the range to (max) 20mA, anything over that is high-current mode. This model has max 5A output"""
        if state:
            return self.set_measure_current_max(0.020)
        return self.set_measure_current_max(5.0)

    def query_low_current_mode(self):
        """Returns boolean indicating whether we are in low or high current mode"""
        max_current = self.query_measure_current_max()
        if max_current <= 0.020:
            return True
        return False

    def measure_current_autorange(self, extra_params=""):
        """Measures the current, then make sure we are running on the correct measurement range and if not switch range and measure again"""
        ret = self.measure_current(extra_params)
        if abs(ret) < 0.020:
            # We need to be in low-current mode
            if not self.query_low_current_mode():
                # Enter low current mode and measure again
                self.set_low_current_mode(True)
                return self.measure_current(extra_params)
            else:
                return ret
        # We need to be in high-current mode
        if self.query_low_current_mode():
            # Switch mode and measure again
            self.set_low_current_mode(False)
            return self.measure_current(extra_params)
        return ret

    def set_remote_mode(self, state=True):
        """RS232 only, prevent accidental button mashing on the fron panel, this switches between SYSTem:REMote and SYSTem:LOCal according to state, this overrides previous value set with set_rwlock"""
        from scpi.transports import rs232 
        if not isinstance(self.scpi.transport, rs232):
            from exceptions import RuntimeError
            raise RuntimeError("Only usable with RS232 transports")
        if state:
            return self.scpi.send_command("SYST:REM", False)
        return self.scpi.send_command("SYST:LOC", False)

    def set_rwlock(self, state=True):
        """RS232 only, prevent *any* button mashing on the fron panel, this switches between SYSTem:RWLock and SYSTem:LOCal according to state, this overrides previous value set with set_remote_mode"""
        from transports import rs232 as serial_transport
        if not isinstance(self.scpi.transport, serial_transport.transports_rs232):
            from exceptions import RuntimeError
            raise RuntimeError("Only usable with RS232 transports")
        if state:
            return self.scpi.send_command("SYST:RWL", False)
        return self.scpi.send_command("SYST:LOC", False)

    def display_on(self, state=True):
        """Sets display on/off"""
        if state:
            return self.scpi.send_command("DISP ON", False)
        return self.scpi.send_command("DISP OFF", False)

    def set_display_mode(self, mode):
        """Set the display mode, valied values are NORM and TEXT"""
        mode = mode.upper()
        if not mode in ( 'NORM', 'TEXT' ):
            raise RuntimeError("Invalid mode %s, valid ones are NORM and TEXT" % mode)
        return self.scpi.send_command("DISP:MODE %s" % mode, False)

    def set_display_text(self, text):
        """Used to display text on the display, max 14 characters, NOTE: does *not* set display mode, you need to do it yourself"""
        if len(text) > 14:
            raise RuntimeError("Max text length is 14 characters")
        if (    '"' in text
            and "'" in text):
            raise RuntimeError("Text may only contain either single or double quotes, not both")
        if '"' in text:
            return self.scpi.send_command("DISP:TEXT '%s'" % text, False)
        return self.scpi.send_command('DISP:TEXT  "%s"' % text, False)
        

def rs232(port, **kwargs):
    """Quick helper to connect via RS232 port"""
    # TODO: figure out why I can't communicate with rtscts enabled (try dsrdtr as well)
    import serial as pyserial
    from scpi.transports import rs232 as serial_transport
    serial_port = pyserial.Serial(port, 9600, timeout=0, **kwargs)
    transport = serial_transport(serial_port)
    dev = hp6632b(transport)
    return dev




"""SCPI module specific errors"""

from exceptions import RuntimeError


class TimeoutError(RuntimeError):
    def __init__(self, command, time, *args, **kwargs):
        self.command = command
        self.time = time
        super(TimeoutError, self).__init__(str(self), *args, **kwargs)

    def __str__(self):
        return "'%s' timed out after %f seconds" % (self.command, self.time)

class CommandError(RuntimeError):
    def __init__(self, command, code, message, *args, **kwargs):
        self.command = command
        self.code = code
        self.message = message
        super(CommandError, self).__init__(str(self), *args, **kwargs)

    def __str__(self):
        return "'%s' returned error %d: %s" % (self.command, self.code, self.message)




from distutils.core import setup
import subprocess

git_version = 'UNKNOWN'
try:
    git_version = str(subprocess.check_output(['git', 'rev-parse', '--verify', '--short', 'HEAD'])).strip()
except subprocess.CalledProcessError,e:
    print "Got error when trying to read git version: %s" % e

setup(
    name='scpi',
    #version='0.6.4dev-%s' % git_version,
    version='0.6.4',
    author='Eero "rambo" af Heurlin',
    author_email='rambo@iki.fi',
    packages=['scpi',],
    license='GNU LGPL',
    long_description=open('README.md').read(),
    description='Implement SCPI in pure Python',
    install_requires=[
        'pyserial>=2.7',
    ],
    url='https://github.com/rambo/python-scpi',
)



"""Baseclass for all the transports, if common methods are needed they will be defined here

All transports must define certain basic methods (check all the raise NotImplementedError)
"""

from exceptions import NotImplementedError
from exceptions import RuntimeError


class transports_base(object):
    def __init__(self):
        """Initializes a transport"""
        pass
    
    def quit(self, command):
        """Must shutdown all background threads (if any)"""
        raise NotImplementedError()

    def set_message_callback(self, callback):
        self.message_received = callback

    def send_command(self, command):
        """Sends a complete command to the device, line termination etc is handled by the transport"""
        raise NotImplementedError()

    def message_received(self, message):
        """Default message callback raises error"""
        raise RuntimeError("Message callback not set")

    def incoming_data(self):
        """Check whether we still have inbound data, must return boolean"""
        raise NotImplementedError()

    def abort_command(self):
        """Send the "device clear" command to abort a running command"""
        raise NotImplementedError()
